from typing import List, Optional

import numpy as np
import pandas as pd
from tensorboard.backend.event_processing import event_accumulator

from matplotlib import pyplot as plt

from .pandas import SeriesInterpolationLinearIndex


class TensorboardData:
    def __init__(self, events: event_accumulator.EventAccumulator):
        self.events = events
        self.events.Reload()

    def get_series(self, tag: str, smoothing_factor: float = 0.0) -> pd.Series:
        """
        Gets the (smoothed) pandas Series for a specific tensorboard tag.

        :param tag: the tensorboard tag
        :param smoothing_factor: the smoothing factor between 0 and 1 which determines the relative importance of past values.
            0: no smoothing
            1: maximum smoothing (all values will be equal to the first value)
        :return: the pandas series with the step as the index
        """
        if not 0 <= smoothing_factor <= 1:
            raise ValueError("Smoothing factor must be between 0 and 1")

        try:
            scalar_events = self.events.Scalars(tag)
        except KeyError:
            raise KeyError(f"Tag '{tag}' not found in tensorboard events")

        steps = [event.step for event in scalar_events]
        values = [event.value for event in scalar_events]

        if smoothing_factor > 0:
            smoothed_values = []
            last = values[0]
            for value in values:
                last = smoothing_factor * last + (1 - smoothing_factor) * value
                smoothed_values.append(last)
            values = smoothed_values

        return pd.Series(values, index=steps, name=tag)

    def get_tags(self) -> List[str]:
        """
        Get list of available scalar tags in the events.

        :return: list of tag names
        """
        return self.events.Tags()['scalars']

    def get_data_frame(self, tags: Optional[List[str]] = None, smoothing_factor: float = 0.0) -> pd.DataFrame:
        """
        Gets multiple series as a DataFrame.

        :param tags: the list of tensorboard tags to consider; if None, use all
        :param smoothing_factor: smoothing factor to apply to all series
        :return: DataFrame with steps as index and tags as columns
        """
        if tags is None:
            tags = self.get_tags()
        series_dict = {}
        for tag in tags:
            series = self.get_series(tag, smoothing_factor)
            series_dict[series.name] = series

        return pd.DataFrame(series_dict)


class TensorboardSeriesComparison:
    def __init__(self, tb_reference: TensorboardData, tb_current: TensorboardData,
            tag: str, index_start: int, index_end: int):
        s_ref = tb_reference.get_series(tag)
        s_cur = tb_current.get_series(tag)

        interp = SeriesInterpolationLinearIndex(ffill=True, bfill=True)
        s_ref, s_cur = interp.interpolate_all_with_combined_index([s_ref, s_cur])

        self.s_ref = s_ref.loc[index_start:index_end]
        self.s_cur = s_cur.loc[index_start:index_end]

    def mean_relative_difference(self):
        """
        Computes the difference between the current series and the reference series, relative to the reference,
        e.g. if the current series is on average 105% of the reference series (5% relative difference), then
        the value will be 0.05.
        Since we divide by the absolute value of the reference, this also works for negative cases, i.e.
        if the reference series value is -0.10 and the current series value is -0.08, then the relative
        difference is 0.2 (20%).

        :return: the mean relative difference
        """
        diff = self.s_cur - self.s_ref
        diff_rel = diff / abs(self.s_ref)
        return np.mean(diff_rel)

    def plot_series(self, show=False) -> plt.Figure:
        fig = plt.figure()
        self.s_ref.plot()
        self.s_cur.plot()
        plt.title(self.s_ref.name)
        if show:
            plt.show()
        return fig



import logging
from dataclasses import dataclass
import subprocess
from typing import Optional

from .string import ToStringMixin

log = logging.getLogger(__name__)


@dataclass
class GitStatus(ToStringMixin):
    commit: str
    has_unstaged_changes: bool
    has_staged_uncommitted_changes: bool
    has_untracked_files: bool

    @property
    def is_clean(self) -> bool:
        return not (self.has_unstaged_changes or
                    self.has_staged_uncommitted_changes or
                    self.has_untracked_files)


def git_status(log_error: bool = True) -> Optional[GitStatus]:
    """
    Gets the git status of the current repository.

    :param log_error: whether to log an error if the git status cannot be determined
    :return: the git status, or None if it cannot be determined
    """
    try:
        commit_hash = subprocess.check_output(['git', 'rev-parse', 'HEAD']).decode('ascii').strip()
        unstaged = bool(subprocess.check_output(['git', 'diff', '--name-only']).decode('ascii').strip())
        staged = bool(subprocess.check_output(['git', 'diff', '--staged', '--name-only']).decode('ascii').strip())
        untracked = bool(subprocess.check_output(['git', 'ls-files', '--others', '--exclude-standard']).decode('ascii').strip())
        return GitStatus(
            commit=commit_hash,
            has_unstaged_changes=unstaged,
            has_staged_uncommitted_changes=staged,
            has_untracked_files=untracked
        )
    except Exception as e:
        if log_error:
            log.error("Error determining Git status", exc_info=e)
        return None



import io
import logging
import os
from typing import Sequence, Optional, Tuple, List, Any, TYPE_CHECKING

if TYPE_CHECKING:
    from matplotlib import pyplot as plt
    import pandas as pd

log = logging.getLogger(__name__)


class ResultWriter:
    log = log.getChild(__qualname__)

    def __init__(self, result_dir, filename_prefix="", enabled: bool = True, close_figures: bool = False):
        """
        :param result_dir:
        :param filename_prefix:
        :param enabled: whether the result writer is enabled; if it is not, it will create neither files nor directories
        :param close_figures: whether to close figures that are passed by default
        """
        self.result_dir = result_dir
        self.filename_prefix = filename_prefix
        self.enabled = enabled
        self.close_figures_default = close_figures
        if self.enabled:
            os.makedirs(result_dir, exist_ok=True)

    def child_with_added_prefix(self, prefix: str) -> "ResultWriter":
        """
        Creates a derived result writer with an added prefix, i.e. the given prefix is appended to this
        result writer's prefix

        :param prefix: the prefix to append
        :return: a new writer instance
        """
        return ResultWriter(self.result_dir, filename_prefix=self.filename_prefix + prefix, enabled=self.enabled,
            close_figures=self.close_figures_default)

    def child_for_subdirectory(self, dir_name: str) -> "ResultWriter":
        result_dir = os.path.join(self.result_dir, dir_name)
        return ResultWriter(result_dir, filename_prefix=self.filename_prefix, enabled=self.enabled,
            close_figures=self.close_figures_default)

    def path(self, filename_suffix: str, extension_to_add=None, valid_other_extensions: Optional[Sequence[str]] = None) -> str:
        """
        :param filename_suffix: the suffix to add (which may or may not already include a file extension)
        :param extension_to_add: if not None, the file extension to add (without the leading ".") unless
            the extension to add or one of the extenions in valid_extensions is already present
        :param valid_other_extensions: a sequence of valid other extensions (without the "."), only
            relevant if extensionToAdd is specified
        :return: the full path
        """
        # replace forbidden characters
        filename_suffix = filename_suffix.replace(">=", "gte").replace(">", "gt")

        if extension_to_add is not None:
            add_ext = True
            valid_extensions = set(valid_other_extensions) if valid_other_extensions is not None else set()
            valid_extensions.add(extension_to_add)
            if valid_extensions is not None:
                for ext in valid_extensions:
                    if filename_suffix.endswith("." + ext):
                        add_ext = False
                        break
            if add_ext:
                filename_suffix += "." + extension_to_add
        path = os.path.join(self.result_dir, f"{self.filename_prefix}{filename_suffix}")
        return path

    def write_text_file(self, filename_suffix: str, content: str):
        p = self.path(filename_suffix, extension_to_add="txt")
        if self.enabled:
            self.log.info(f"Saving text file {p}")
            with open(p, "w") as f:
                f.write(content)
        return p

    def write_text_file_lines(self, filename_suffix: str, lines: List[str]):
        p = self.path(filename_suffix, extension_to_add="txt")
        if self.enabled:
            self.log.info(f"Saving text file {p}")
            write_text_file_lines(lines, p)
        return p

    def write_data_frame_text_file(self, filename_suffix: str, df: "pd.DataFrame"):
        p = self.path(filename_suffix, extension_to_add="df.txt", valid_other_extensions="txt")
        if self.enabled:
            self.log.info(f"Saving data frame text file {p}")
            with open(p, "w") as f:
                f.write(df.to_string())
        return p

    def write_data_frame_csv_file(self, filename_suffix: str, df: "pd.DataFrame", index=True, header=True):
        p = self.path(filename_suffix, extension_to_add="csv")
        if self.enabled:
            self.log.info(f"Saving data frame CSV file {p}")
            df.to_csv(p, index=index, header=header)
        return p

    def write_data_frame_excel_file(self, filename_suffix: str, df: "pd.DataFrame", index=True, header=True, sheet_name="Sheet1"):
        """
        Saves the given data frame to an Excel spreadsheet.

        Requires that an appropriate engine be installed, e.g. openpyxl.

        :param filename_suffix: the filename suffix (or full basename for the case where no prefix was specified at construction),
            to which the extension .xlsx will be added by default if no extension is given
        :param df: the data frame to save
        :param index: whether to save the index
        :param header: whether to save the header row
        :param sheet_name: the name of the spreadsheet
        :return: the path to the saved file
        """
        p = self.path(filename_suffix, extension_to_add="xlsx")
        if self.enabled:
            self.log.info(f"Saving data frame Excel file {p}")
            df.to_excel(p, index=index, header=header, sheet_name=sheet_name)
        return p

    def write_figure(self, filename_suffix: str, fig: "plt.Figure", close_figure: Optional[bool] = None):
        """
        :param filename_suffix: the filename suffix, which may or may not include a file extension, valid extensions being {"png", "jpg"}
        :param fig: the figure to save
        :param close_figure: whether to close the figure after having saved it; if None, use default passed at construction
        :return: the path to the file that was written (or would have been written if the writer was enabled)
        """
        from matplotlib import pyplot as plt
        p = self.path(filename_suffix, extension_to_add="png", valid_other_extensions=("jpg",))
        if self.enabled:
            self.log.info(f"Saving figure {p}")
            fig.savefig(p, bbox_inches="tight")
            must_close_figure = close_figure if close_figure is not None else self.close_figures_default
            if must_close_figure:
                plt.close(fig)
        return p

    def write_figures(self, figures: Sequence[Tuple[str, "plt.Figure"]], close_figures=False):
        for name, fig in figures:
            self.write_figure(name, fig, close_figure=close_figures)

    def write_pickle(self, filename_suffix: str, obj: Any):
        from .pickle import dump_pickle
        p = self.path(filename_suffix, extension_to_add="pickle")
        if self.enabled:
            self.log.info(f"Saving pickle {p}")
            dump_pickle(obj, p)
        return p


def write_text_file_lines(lines: List[str], path):
    """
    :param lines: the lines to write (without a trailing newline, which will be added)
    :param path: the path of the text file to write to
    """
    with open(path, "w") as f:
        for line in lines:
            f.write(line)
            f.write("\n")


def read_text_file_lines(path, strip=True, skip_empty=True) -> List[str]:
    """
    :param path: the path of the text file to read from
    :param strip: whether to strip each line, removing whitespace/newline characters
    :param skip_empty: whether to skip any lines that are empty (after stripping)
    :return: the list of lines
    """
    lines = []
    with open(path, "r") as f:
        for line in f.readlines():
            if strip:
                line = line.strip()
            if not skip_empty or line != "":
                lines.append(line)
    return lines


def is_s3_path(path: str):
    return path.startswith("s3://")


class S3Object:
    def __init__(self, path):
        assert is_s3_path(path)
        self.path = path
        self.bucket, self.object = self.path[5:].split("/", 1)

    class OutputFile:
        def __init__(self, s3_object: "S3Object"):
            self.s3Object = s3_object
            self.buffer = io.BytesIO()

        def write(self, obj: bytes):
            self.buffer.write(obj)

        def __enter__(self):
            return self

        def __exit__(self, exc_type, exc_val, exc_tb):
            self.s3Object.put(self.buffer.getvalue())

    def get_file_content(self):
        return self._get_s3_object().get()['Body'].read()

    def open_file(self, mode):
        assert mode in ("wb", "rb")
        if mode == "rb":
            content = self.get_file_content()
            return io.BytesIO(content)

        elif mode == "wb":
            return self.OutputFile(self)

        else:
            raise ValueError(mode)

    def put(self, obj: bytes):
        self._get_s3_object().put(Body=obj)

    def _get_s3_object(self):
        import boto3
        session = boto3.session.Session(profile_name=os.getenv("AWS_PROFILE"))
        s3 = session.resource("s3")
        return s3.Bucket(self.bucket).Object(self.object)


def create_path(root: str, *path_elems: str, is_dir: bool, make_dirs: bool = False) -> str:
    path = os.path.join(root, *path_elems)
    if make_dirs:
        dir_path = path if is_dir else os.path.dirname(path)
        os.makedirs(dir_path, exist_ok=True)
    return path


def create_file_path(root, *path_elems, make_dirs: bool = False) -> str:
    return create_path(root, *path_elems, is_dir=False, make_dirs=make_dirs)


def create_dir_path(root, *path_elems, make_dirs: bool = False) -> str:
    return create_path(root, *path_elems, is_dir=True, make_dirs=make_dirs)


def filename_compatible(fn: str, replacement: str = "-") -> str:
    """
    Converts the given string into a string that can be used in a filename

    :param fn: original filename, which may or may not be compatible with common filenames
    :param replacement: default character to use as a replacement
    :return: adapted filename
    """
    # Replace prohibited characters
    # Windows prohibits: \ / : * ? " < > |
    # Unix/Linux prohibits: /
    # macOS prohibits: / and :
    replacements = {
        ">": "gt",
        "<": "lt",
        "/": replacement,
        "\\": replacement,
        ":": replacement,
        "*": replacement,
        "?": replacement,
        "\"": replacement,
        "|": replacement
    }

    result = fn
    for char, repl in replacements.items():
        result = result.replace(char, repl)

    return result



import enum
import logging
import pickle
import pandas as pd

from .cache import PersistentKeyValueCache, DelayedUpdateHook

log = logging.getLogger(__name__)


class MySQLPersistentKeyValueCache(PersistentKeyValueCache):
    """
    Can cache arbitrary values in a MySQL database.
    The keys are always strings at the database level, i.e. if a key is not a string, it is converted to a string using str().
    """

    class ValueType(enum.Enum):
        """
        The value type to use within the MySQL database.
        Note that the binary BLOB types can be used for all Python types that can be pickled, so the lack
        of specific types (e.g. for strings) is not a problem.
        """
        # enum values are (SQL data type, isCachedValuePickled)
        DOUBLE = ("DOUBLE", False)
        BLOB = ("BLOB", True)
        """
        for Python data types whose pickled representation is up to 64 KB
        """
        MEDIUMBLOB = ("MEDIUMBLOB", True)
        """
        for Python data types whose pickled representation is up to 16 MB
        """

    def __init__(self, host: str, db: str, user: str, pw: str, value_type: ValueType, table_name="cache",
        connect_params: dict | None = None, in_memory=False, max_key_length: int = 255, port=3306):
        """
        :param host:
        :param db:
        :param user:
        :param pw:
        :param value_type: the type of value to store in the cache
        :param table_name:
        :param connect_params: additional parameters to pass to the pymysql.connect() function (e.g. ssl, etc.)
        :param in_memory:
        :param max_key_length: maximal length of the cache key string (keys are always strings) stored in the DB
            (i.e. the MySQL type is VARCHAR[max_key_length])
        :param port: the MySQL server port to connect to
        """
        import pymysql
        if connect_params is None:
            connect_params = {}
        self._connect = lambda: pymysql.connect(host=host, database=db, user=user, password=pw, port=port, autocommit=True,
            **connect_params)
        self._conn = self._connect()
        self.table_name = table_name
        self.max_key_length = max_key_length

        cache_value_sql_type, self.is_cache_value_pickled = value_type.value

        cursor = self._conn.cursor()
        cursor.execute(f"SHOW TABLES;")
        if table_name not in [r[0] for r in cursor.fetchall()]:
            log.debug(f"Creating table {table_name}")
            cursor.execute(f"CREATE TABLE {table_name} (cache_key VARCHAR({self.max_key_length}) PRIMARY KEY, "
                           f"cache_value {cache_value_sql_type});")
        cursor.close()

        self._in_memory_df = None if not in_memory else self._load_table_to_data_frame()

    def _cursor(self):
        try:
            self._conn.ping(reconnect=True)
        except Exception as e:
            log.error(f"Error while pinging MySQL server: {e}; Reconnecting ...")
            self._conn = self._connect()
        return self._conn.cursor()

    def _load_table_to_data_frame(self):
        df = pd.read_sql(f"SELECT * FROM {self.table_name};", con=self._conn, index_col="cache_key")
        if self.is_cache_value_pickled:
            df["cache_value"] = df["cache_value"].apply(pickle.loads)
        return df

    def set(self, key, value):
        key = str(key)
        if len(key) > self.max_key_length:
            raise ValueError(f"Key too long, maximal key length is {self.max_key_length}")
        cursor = self._cursor()
        cursor.execute(f"SELECT COUNT(*) FROM {self.table_name} WHERE cache_key=%s", (key,))
        stored_value = pickle.dumps(value) if self.is_cache_value_pickled else value
        if cursor.fetchone()[0] == 0:
            from pymysql.err import IntegrityError
            try:
                cursor.execute(f"INSERT INTO {self.table_name} (cache_key, cache_value) VALUES (%s, %s)",
                    (key, stored_value))
            except IntegrityError as e:
                if e.args[0] == 1062:  # Duplicate entry
                    # This can only happen when the user is inserting the same value almost simultaneously (race condition)
                    args = list(e.args)
                    args[1] = f"{args[1]}; The duplicate entry is due to quasi-simultaneous insertions for the same key; " \
                              "Check your application logic!"
                    raise IntegrityError(*args)
                else:
                    raise
        else:
            cursor.execute(f"UPDATE {self.table_name} SET cache_value=%s WHERE cache_key=%s", (stored_value, key))
        cursor.close()
        if self._in_memory_df is not None:
            self._in_memory_df["cache_value"][str(key)] = value

    def get(self, key):
        value = self._get_from_in_memory_df(key)
        if value is None:
            value = self._get_from_table(key)
        return value

    def _get_from_table(self, key):
        cursor = self._cursor()
        cursor.execute(f"SELECT cache_value FROM {self.table_name} WHERE cache_key=%s", (str(key),))
        row = cursor.fetchone()
        cursor.close()
        if row is None:
            return None
        stored_value = row[0]
        value = pickle.loads(stored_value) if self.is_cache_value_pickled else stored_value
        return value

    def _get_from_in_memory_df(self, key):
        if self._in_memory_df is None:
            return None
        try:
            return self._in_memory_df["cache_value"][str(key)]
        except Exception as e:
            log.debug(f"Unable to load value for key {str(key)} from in-memory dataframe: {e}")
            return None



from abc import ABC, abstractmethod
from enum import Enum
import logging
from typing import Callable, Union, TypeVar, Generic, Sequence, List, Tuple, Iterable, Dict, Hashable, Optional

import numpy as np

from .util import LogTime
from .util.pickle import setstate
from .util.string import list_string, ToStringMixin

T = TypeVar("T")

log = logging.getLogger(__name__)


class Vectoriser(Generic[T], ToStringMixin):
    """
    A vectoriser represents a method for the conversion of instances of some type T into
    vectors, i.e. one-dimensional (numeric) arrays, or (in the special case of a 1D vector) scalars
    """

    log = log.getChild(__qualname__)

    def __init__(self, f: Callable[[T], Union[float, np.ndarray, list]], transformer=None, is_fitted=False):
        """
        :param f: the function which maps from an instance of T to an array/list/scalar
        :param transformer: an optional transformer (e.g. instance of one of the classes in sklearn.preprocessing)
            which can be used to transform/normalise the generated arrays
        :param is_fitted: whether the vectoriser (and therefore the given transformer) is assumed to be fitted already
        """
        self._fn = f
        self.transformer = transformer
        self._resultType = None
        self.name = None
        self._is_fitted = is_fitted

    def __setstate__(self, state):
        new_default_properties = {
            "_is_fitted": True  # we assume that any persisted objects have been fitted
        }
        setstate(Vectoriser, self, state, new_optional_properties=["_resultType", "name"], renamed_properties={"f": "_fn"},
            new_default_properties=new_default_properties)

    def _tostring_exclude_private(self) -> bool:
        return True

    def is_fitted(self) -> bool:
        return self._is_fitted

    def set_name(self, name):
        self.name = name

    def get_name(self):
        """
        :return: the name of this feature generator, which may be a default name if the name has not been set. Note that feature generators
            created by a FeatureGeneratorFactory always get the name with which the generator factory was registered.
        """
        if self.name is None:
            return f"{self.__class__.__name__}-{id(self)}"
        return self.name

    def fit(self, items: Iterable[T]):
        if self.transformer is not None:
            values = [self._f(item) for item in items]
            self.transformer.fit(np.array(values))
        self._is_fitted = True

    def _f(self, x) -> np.array:
        y = self._fn(x)

        if self._resultType is None:
            self._resultType = self.ResultType.from_value(y)

        if self._resultType == self.ResultType.LIST:
            y = np.array(y)
        elif self._resultType == self.ResultType.SCALAR:
            y = np.array([y])

        return y

    def apply(self, item: T, transform=True) -> np.array:
        """
        :param item: the item to be vectorised
        :param transform: whether to apply this instance's transformer (if any)
        :return: a vector
        """
        value = self._f(item)
        if self.transformer is not None and transform:
            value = self.transformer.transform([value])[0]
        return value

    def apply_multi(self, items: Iterable[T], transform=True, use_cache=False, verbose=False) -> List[np.array]:
        """
        Applies this vectoriser to multiple items at once.
        Especially for cases where this vectoriser uses a transformer, this method is significantly faster than
        calling apply repeatedly.

        :param items: the items to be vectorised
        :param transform: whether to apply this instance's transformer (if any)
        :param use_cache: whether to apply caching of the value function f given at construction (keeping track of outputs for
            each input object id), which can significantly speed up computation in cases where an items appears more than
            once in the collection of items
        :param verbose: whether to generate log messages
        :return: a list of vectors
        """
        if verbose:
            self.log.info(f"Applying {self}")

        with LogTime("Application", enabled=verbose, logger=self.log):
            if not use_cache:
                compute_value = self._f
            else:
                cache = {}

                def compute_value(x):
                    key = id(x)
                    value = cache.get(key)
                    if value is None:
                        value = self._f(x)
                        cache[key] = value
                    return value

            values = [compute_value(x) for x in items]
            if self.transformer is not None and transform:
                values = self.transformer.transform(values)
            return values

    class ResultType(Enum):
        SCALAR = 0
        LIST = 1
        NUMPY_ARRAY = 2

        @classmethod
        def from_value(cls, y):
            if type(y) == list:
                return cls.LIST
            elif np.isscalar(y):
                return cls.SCALAR
            elif isinstance(y, np.ndarray):
                return cls.NUMPY_ARRAY
            else:
                raise ValueError(f"Received unhandled value of type {type(y)}")


class EmptyVectoriser(Vectoriser):
    def __init__(self):
        super().__init__(self._create_empty_vector)

    # noinspection PyUnusedLocal
    @staticmethod
    def _create_empty_vector(x):
        return np.zeros(0)


class ItemIdentifierProvider(Generic[T], ABC):
    """
    Provides identifiers for sequence items.
    """
    @abstractmethod
    def get_identifier(self, item: T) -> Hashable:
        pass


class SequenceVectoriser(Generic[T], ToStringMixin):
    """
    Supports the application of Vectorisers to sequences of objects of some type T, where each object of type T is
    mapped to a vector (1D array) by the vectorisers.
    A SequenceVectoriser is fitted by fitting the underlying Vectorisers. In order to obtain the instances of T that
    are used for training, we take into consideration the fact that the sequences of T may overlap and thus training
    is performed on the set of unique instances.
    """

    log = log.getChild(__qualname__)

    class FittingMode(Enum):
        """
        Determines how the individual vectorisers are fitted based on several sequences of objects of type T that are given.
        If NONE, no fitting is performed, otherwise the mode determines how a single sequence of objects of type T for fitting
        is obtained from the collection of sequences: either by forming the set of unique objects from the sequences (UNIQUE)
        """
        NONE = "none"  # no fitting is performed
        UNIQUE = "unique"  # use collection of unique items
        CONCAT = "concat"  # use collection obtained by concatenating all sequences using numpy.concatenate

    def __init__(self, vectorisers: Union[Sequence[Vectoriser[T]], Vectoriser[T]],
            fitting_mode: FittingMode = FittingMode.UNIQUE,
            unique_id_provider: Optional[ItemIdentifierProvider] = None,
            refit_vectorisers: bool = True):
        """
        :param vectorisers: zero or more vectorisers that are to be applied. If more than one vectoriser is supplied,
            vectors are generated from input instances of type T by concatenating the results of the vectorisers in
            the order the vectorisers are given.
        :param fitting_mode: the fitting mode for vectorisers. If `NONE`, no fitting takes place.
            If `UNIQUE`, fit vectorisers on unique set of items of type T. By default, uniqueness is determined based
            on Python object identity. If a custom mechanisms for determining an item's identity is desired,
            pass `unique_id_retriever`.
            If `CONCAT`, fit vectorisers based on all items of type T, concatenating them to a single sequence.
        :param unique_id_provider: an object used to determine item identities when using fitting mode `UNIQUE`.
        :param refit_vectorisers: whether any vectorisers that have previously been fitted shall be
            fitted once more when this sequence vectoriser is fitted. Set this to false if you are reusing vectorisers
            that are also part of another sequence vectoriser that will be fitted/has been fitted before this
            sequence vectoriser. This can be useful, in particular, in encoder-decoders where the target features
            are partly the same as the history sequence features, and we want to reuse the latter and their
            fitted transformers for the target features.
        """
        self.fittingMode = fitting_mode
        self.uniqueIdProvider = unique_id_provider
        self.refitVectorisers = refit_vectorisers
        if isinstance(vectorisers, Vectoriser):
            self.vectorisers = [vectorisers]
        else:
            self.vectorisers = vectorisers
        if len(self.vectorisers) == 0:
            self.vectorisers = [EmptyVectoriser()]

    def __setstate__(self, state):
        state["fittingMode"] = state.get("fittingMode", self.FittingMode.UNIQUE)
        setstate(SequenceVectoriser, self, state, new_optional_properties=["uniqueIdProvider"],
            new_default_properties={"refitVectorisers": True})

    def fit(self, data: Iterable[Sequence[T]]):
        log.debug(f"Fitting {self}")

        if self.fittingMode == self.FittingMode.NONE:
            return

        if not self.refitVectorisers:
            if all(v.is_fitted() for v in self.vectorisers):
                log.debug(f"No vectorisers to be fitted; all contained vectorisers have previously been fitted")
                return

        # obtain items for fitting
        if self.fittingMode == self.FittingMode.UNIQUE:
            if self.uniqueIdProvider is None:
                items = set()
                for seq in data:
                    items.update(seq)
            else:
                items = []
                identifiers = set()
                for seq in data:
                    for item in seq:
                        identifier = self.uniqueIdProvider.get_identifier(item)
                        if identifier not in identifiers:
                            identifiers.add(identifier)
                            items.append(item)
        elif self.fittingMode == self.FittingMode.CONCAT:
            items = np.concatenate(data)  # type: ignore
        else:
            raise ValueError(self.fittingMode)

        for v in self.vectorisers:
            if self.refitVectorisers or not v.is_fitted():
                log.debug(f"Fitting {v}")
                v.fit(items)

    def apply(self, seq: Sequence[T], transform=True) -> List[np.array]:
        """
        Applies vectorisation to the given sequence of objects

        :param seq: the sequence to vectorise
        :param transform: whether to apply any post-vectorisation transformers
        :return:
        """
        vectors_list = []
        for item in seq:
            vectors = [vec.apply(item, transform=transform) for vec in self.vectorisers]
            conc = np.concatenate(vectors, axis=0)
            vectors_list.append(conc)
        return vectors_list

    def apply_multi(self, sequences: Iterable[Sequence[T]], use_cache=False, verbose=False) -> Tuple[List[List[np.array]], List[int]]:
        """
        Applies this vectoriser to multiple sequences of objects of type T, where each sequence is mapped to a sequence
        of 1D arrays.
        This method can be significantly faster than multiple applications of apply, especially in cases where the vectorisers
        use transformers.

        :param sequences: the sequences to vectorise
        :param use_cache: whether to apply caching of the value functions of contained vectorisers (keeping track of outputs for
            each input object id), which can significantly speed up computation in cases where the given sequences contain individual
            items more than once
        :param verbose: whether to generate log messages
        :return: a pair (vl, l) where vl is a list of lists of vectors/arrays and l is a list of integers containing the lengths
            of the sequences
        """
        if verbose:
            self.log.info(f"Applying {self} (useCache={use_cache})")

        lengths = [len(s) for s in sequences]

        if verbose:
            self.log.info("Generating combined sequence")
        combined_seq = []
        for seq in sequences:
            combined_seq.extend(seq)

        individual_vectoriser_results = [vectoriser.apply_multi(combined_seq, use_cache=use_cache, verbose=verbose)
            for vectoriser in self.vectorisers]
        conc_vectors = [np.concatenate(x, axis=0) for x in zip(*individual_vectoriser_results)]

        vector_sequences = []
        idx_start = 0
        for l in lengths:
            vector_sequences.append(conc_vectors[idx_start:idx_start+l])
            idx_start += l

        return vector_sequences, lengths

    def apply_multi_with_padding(self, sequences: Sequence[Sequence[T]], use_cache=False, verbose=False) \
            -> Tuple[List[List[np.array]], List[int]]:
        """
        Applies this vectoriser to multiple sequences of objects of type T, where each sequence is mapped to a sequence
        of 1D arrays.
        Sequences are allowed to vary in length. for shorter sequences, 0-vectors are appended until the maximum sequence length
        is reached (padding).

        :param sequences: the sequences to vectorise
        :param use_cache: whether to apply caching of the value functions of contained vectorisers (keeping track of outputs for
            each input object id), which can significantly speed up computation in cases where the given sequences contain individual
            items more than once
        :param verbose: whether to generate log messages
        :return: a pair (vl, l) where vl is a list of lists of vectors/arrays, each list having the same length, and l is a list of
            integers containing the original unpadded lengths of the sequences
        """
        result, lengths = self.apply_multi(sequences, use_cache=use_cache, verbose=verbose)
        if verbose:
            self.log.info("Applying padding")
        max_length = max(lengths)
        dim = len(result[0][0])
        dummy_vec = np.zeros((dim,))
        for seq in result:
            for i in range(max_length - len(seq)):
                seq.append(dummy_vec)
        return result, lengths

    def get_vector_dim(self, seq: Sequence[T]):
        """
        Determines the dimensionality of generated vectors by applying the vectoriser to the given sequence

        :param seq: the sequence
        :return: the number of dimensions in generated output vectors (per item)
        """
        return len(self.apply(seq, transform=False)[0])


class VectoriserRegistry:
    def __init__(self):
        self._factories: Dict[Hashable, Callable[[Callable], Vectoriser]] = {}

    def get_available_vectorisers(self):
        return list(self._factories.keys())

    @staticmethod
    def _name(name: Hashable):
        # for enums, which have .name, use the name only, because it is less problematic to persist
        if hasattr(name, "name"):
            name = name.name
        return name

    def register_factory(self, name: Hashable, factory: Callable[[Callable], Vectoriser],
            additional_names: Optional[Iterable[Hashable]] = None):
        """
        Registers a vectoriser factory which can subsequently be referenced via their name

        :param name: the name (which can, in particular, be a string or an enum item)
        :param factory: the factory, which takes the default transformer factory as an argument
        :param additional_names: (optional) additional names under which to register the factory
        """
        self._register_factory(name, factory)
        if additional_names is not None:
            for n in additional_names:
                self._register_factory(n, factory)

    def _register_factory(self, name: Hashable, factory):
        name = self._name(name)
        if name in self._factories:
            raise ValueError(f"Vectoriser factory for name '{name}' already registered")
        self._factories[name] = factory

    def get_vectoriser(self, name: Hashable, default_transformer_factory: Callable) -> Vectoriser:
        """
        Creates a vectoriser from a name, which must have been previously registered.

        :param name: the name (which can, in particular, be a string or an enum item)
        :param default_transformer_factory: the default transformer factory
        :return: a new vectoriser instance
        """
        name = self._name(name)
        factory = self._factories.get(name)
        if factory is None:
            raise ValueError(f"No vectoriser factory registered for name '{name}': known names: {list_string(self._factories.keys())}. "
                             f"Register the factory first.")
        instance = factory(default_transformer_factory)
        instance.set_name(name)
        return instance

    def get_vectorisers(self, names: List[Hashable], default_transformer_factory: Callable) -> List[Vectoriser]:
        return [self.get_vectoriser(name, default_transformer_factory) for name in names]



import logging
from typing import Optional, List, Union, Dict, Hashable

import pandas as pd
import torch

from ... import TorchVectorClassificationModel, VectorTorchModel, ClassificationOutputMode
from ....torch import TorchVectorRegressionModel, Tensoriser, NNOptimiserParams, TorchModel
from ....torch.torch_models.seq.seq_modules import EncoderDecoderModule, EncoderFactory, DecoderFactory
from ....vectoriser import SequenceVectoriser

log = logging.getLogger(__name__)


class EncoderDecoderVectorRegressionModel(TorchVectorRegressionModel):
    """
    A highly general encoder-decoder sequence model, which encodes a sequence of history items
    and uses the encoding to make predictions for one or more target sequence items.
    History and target sequences are converted to vectors via vectorisers.
    """
    def __init__(self, cuda: bool,
            history_sequence_column_name: str,
            history_sequence_vectoriser: SequenceVectoriser,
            history_sequence_variable_length: bool,
            target_sequence_column_name: str,
            target_sequence_vectoriser: SequenceVectoriser,
            latent_dim: int,
            encoder_factory: EncoderFactory,
            decoder_factory: DecoderFactory,
            nn_optimiser_params: Optional[NNOptimiserParams] = None):
        """
        :param cuda: whether to use a CUDA device
        :param history_sequence_column_name: the name of the data frame input column which contains the history sequences to be encoded.
            The column must contain a sequence of items that can be converted to vectors via the `history_sequence_vectorizer`
        :param history_sequence_vectoriser: a vectorizer which converts history sequence items to vectors
        :param history_sequence_variable_length: whether history sequences can be of variable length
        :param target_sequence_column_name: the column containing the target item sequence; Note that the column must
            contain sequences even if there is but a single target item for which predictions shall be made.
            In such cases, simply use a column that contains lists with a single item each.
        :param target_sequence_vectoriser: the vectoriser for the generation of feature vectors for the target
            items.
        :param latent_dim: the number of latent dimensions to be used by the encoder
        :param encoder_factory: a factory for the creation of the encoder, which takes sequence items from the history
            and encodes them into vectors of dimension `latent_dim`
        :param decoder_factory: a factory for the creation of the decoder component, which takes a latent vector produced by the
            encoder and (a sequence of) target features to make predictions
        :param nn_optimiser_params: the optimiser parameters
        """
        super().__init__(self._create_model, nn_optimiser_params=nn_optimiser_params)
        self.history_sequence_variable_length = history_sequence_variable_length
        self.latent_dim = latent_dim
        self.cuda = cuda
        self.decoder_factory = decoder_factory
        self.encoder_factory = encoder_factory
        self.sequenceVectoriser = history_sequence_vectoriser
        self.history_sequence_dim_per_item: Optional[int] = None
        self.target_feature_dim: Optional[int] = None
        self.with_input_tensoriser(self.InputTensoriser(history_sequence_column_name, history_sequence_vectoriser,
            target_sequence_column_name, target_sequence_vectoriser))

    def _create_model(self):
        return self.EncoderDecoderModel(self)

    class InputTensoriser(Tensoriser):
        def __init__(self,
                history_sequence_column_name: str,
                history_sequence_vectoriser: SequenceVectoriser,
                target_sequence_column_name: str,
                target_sequence_vectoriser: SequenceVectoriser):
            self.history_sequence_column_name = history_sequence_column_name
            self.history_sequence_vectoriser = history_sequence_vectoriser
            self.target_sequence_column_name = target_sequence_column_name
            self.target_sequence_vectoriser = target_sequence_vectoriser

        def fit(self, df: pd.DataFrame, model=None):
            model: "EncoderDecoderVectorRegressionModel"
            self.history_sequence_vectoriser.fit(df[self.history_sequence_column_name])
            model.history_sequence_dim_per_item = self.history_sequence_vectoriser.get_vector_dim(df[self.history_sequence_column_name].iloc[0])
            self.target_sequence_vectoriser.fit(df[self.target_sequence_column_name])
            model.target_feature_dim = self.target_sequence_vectoriser.get_vector_dim(df[self.target_sequence_column_name].iloc[0])

        def _tensorise(self, df: pd.DataFrame) -> Union[torch.Tensor, List[torch.Tensor]]:
            log.debug(f"Applying {self} to data frame of length {len(df)} ...")
            history = df[self.history_sequence_column_name]
            history_sequences, history_sequence_lengths = self.history_sequence_vectoriser.apply_multi_with_padding(history)
            targets = df[self.target_sequence_column_name]
            target_sequences, target_sequence_lengths = self.target_sequence_vectoriser.apply_multi_with_padding(targets)
            return [
                torch.tensor(history_sequences).float(),
                torch.tensor(history_sequence_lengths),
                torch.tensor(target_sequences).float(),
                torch.tensor(target_sequence_lengths),
            ]

    class EncoderDecoderModel(TorchModel):
        def __init__(self, parent: "EncoderDecoderVectorRegressionModel"):
            super().__init__(parent.cuda)
            self.parent = parent

        def create_torch_module(self) -> torch.nn.Module:
            latent_dim = self.parent.latent_dim
            target_feature_dim = self.parent.target_feature_dim
            encoder = self.parent.encoder_factory.create_encoder(self.parent.history_sequence_dim_per_item, latent_dim)
            decoder = self.parent.decoder_factory.create_decoder(latent_dim, target_feature_dim)
            return EncoderDecoderModule(encoder, decoder, self.parent.history_sequence_variable_length)


class EncoderDecoderVectorClassificationModel(TorchVectorClassificationModel):
    """
    A highly general encoder-decoder sequence model, which encodes a sequence of history items
    and uses the encoding to make predictions for one or more target sequence items.
    History input sequences are converted to vectors via a vectoriser.
    """
    def __init__(self,
            output_mode: ClassificationOutputMode,
            cuda: bool,
            history_sequence_column_name: str,
            history_sequence_vectoriser: SequenceVectoriser,
            history_sequence_variable_length: bool,
            latent_dim: int,
            encoder_factory: EncoderFactory,
            decoder_factory: DecoderFactory,
            nn_optimiser_params: Optional[NNOptimiserParams] = None,
            class_weights: Optional[Dict[Hashable, float]] = None):
        """
        :param output_mode: the output mode defining the semantics of the outputs produced by the decoder
        :param cuda: whether to use a CUDA device
        :param history_sequence_column_name: the name of the data frame input column which contains the history sequences to be encoded.
            The column must contain a sequence of items that can be converted to vectors via the `history_sequence_vectorizer`
        :param history_sequence_vectoriser: a vectorizer which converts history sequence items to vectors
        :param history_sequence_variable_length: whether history sequences can be of variable length
        :param latent_dim: the number of latent dimensions to be used by the encoder
        :param encoder_factory: a factory for the creation of the encoder, which takes sequence items from the history
            and encodes them into vectors of dimension `latent_dim`
        :param decoder_factory: a factory for the creation of the decoder component, which takes a latent vector produced by the
            encoder and (a sequence of) target features to make predictions
        :param nn_optimiser_params: the optimiser parameters
        :param class_weights: a mapping from class labels to weights (which will be applied to the default loss evaluator,
            provided that it is not overridden in `nn_optimiser_params`)
        """
        super().__init__(output_mode,
            self._create_model,
            nn_optimiser_params=nn_optimiser_params,
            class_weights=class_weights)
        self.history_sequence_variable_length = history_sequence_variable_length
        self.latent_dim = latent_dim
        self.cuda = cuda
        self.decoder_factory = decoder_factory
        self.encoder_factory = encoder_factory
        self.sequenceVectoriser = history_sequence_vectoriser
        self.history_sequence_dim_per_item: Optional[int] = None
        self.with_input_tensoriser(self.InputTensoriser(history_sequence_column_name, history_sequence_vectoriser))

    def _create_model(self):
        return self.EncoderDecoderModel(self)

    class InputTensoriser(Tensoriser):
        def __init__(self,
                history_sequence_column_name: str,
                history_sequence_vectoriser: SequenceVectoriser):
            self.history_sequence_column_name = history_sequence_column_name
            self.history_sequence_vectoriser = history_sequence_vectoriser

        def fit(self, df: pd.DataFrame, model=None):
            model: "EncoderDecoderVectorRegressionModel"
            self.history_sequence_vectoriser.fit(df[self.history_sequence_column_name])
            model.history_sequence_dim_per_item = self.history_sequence_vectoriser.get_vector_dim(df[self.history_sequence_column_name].iloc[0])

        def _tensorise(self, df: pd.DataFrame) -> Union[torch.Tensor, List[torch.Tensor]]:
            log.debug(f"Applying {self} to data frame of length {len(df)} ...")
            history = df[self.history_sequence_column_name]
            history_sequences, history_sequence_lengths = self.history_sequence_vectoriser.apply_multi_with_padding(history)
            return [
                torch.tensor(history_sequences).float(),
                torch.tensor(history_sequence_lengths),
            ]

    class EncoderDecoderModel(VectorTorchModel):
        def __init__(self, parent: "EncoderDecoderVectorClassificationModel"):
            super().__init__(parent.cuda)
            self.parent = parent

        def create_torch_module_for_dims(self, input_dim: int, output_dim: int) -> EncoderDecoderModule:
            latent_dim = self.parent.latent_dim
            encoder = self.parent.encoder_factory.create_encoder(self.parent.history_sequence_dim_per_item, latent_dim)
            decoder = self.parent.decoder_factory.create_decoder(latent_dim, 0, output_dim)
            return EncoderDecoderModule(encoder, decoder, self.parent.history_sequence_variable_length)



# -*- coding: utf-8 -*-
import logging
import math
import queue
from abc import ABC, abstractmethod
from typing import List, Sequence, Iterator, Optional, Union, Tuple

log = logging.getLogger(__name__)


class GreedyAgglomerativeClustering(object):
    """
    An implementation of greedy agglomerative clustering which avoids unnecessary 
    recomputations of merge costs through the management of a priority queue of 
    potential merges.

    Greedy agglomerative clustering works as follows. Starting with an initial
    set of clusters (where each cluster typically contains a single data point),
    the method successively merges the two clusters where the merge cost is lowest (greedy),
    until no further merges are admissible.
    The merge operation is a mutating operation, i.e. the initial clusters are modified.

    To apply the method, the Cluster class must be subclassed, so as to define
    what the cost of a merge in your application shall be and how two clusters can be merged.
    For example, if data points are points in a Cartesian coordinate system, then the merge cost
    can be defined as the minimum or maximum distance among all pairs of points in the two clusters,
    admissibility being determined by a threshold that must not be exceeded;
    the merge operation can simply concatenate lists of data points.
    """
    log = log.getChild(__qualname__)

    class Cluster(ABC):
        """
        Base class for clusters that can be merged via GreedyAgglomerativeClustering
        """
        @abstractmethod
        def merge_cost(self, other) -> float:
            """
            Computes the cost of merging the given cluster with this cluster

            :return: the (non-negative) merge cost or math.inf if a merge is inadmissible"""
            pass

        @abstractmethod
        def merge(self, other):
            """
            Merges the given cluster into this cluster"

            :param other: the cluster that is to be merged into this cluster
            """
            pass

    def __init__(self, clusters: Sequence[Cluster],
            merge_candidate_determination_strategy: "GreedyAgglomerativeClustering.MergeCandidateDeterminationStrategy" = None):
        """
        :param clusters: the initial clusters, which are to be agglomerated into larger clusters
        """
        self.prioritised_merges = queue.PriorityQueue()
        self.wrapped_clusters = []
        for idx, c in enumerate(clusters):
            self.wrapped_clusters.append(GreedyAgglomerativeClustering.WrappedCluster(c, idx, self))

        # initialise merge candidate determination strategy
        if merge_candidate_determination_strategy is None:
            merge_candidate_determination_strategy = self.MergeCandidateDeterminationStrategyDefault()
        merge_candidate_determination_strategy.set_clusterer(self)
        self.mergeCandidateDeterminationStrategy = merge_candidate_determination_strategy
        
    def apply_clustering(self) -> List[Cluster]:
        """
        Applies greedy agglomerative clustering to the clusters given at construction, merging
        clusters until no further merges are admissible

        :return: the list of agglomerated clusters (subset of the original clusters, which may have had other
            clusters merged into them)
        """
        # compute all possible merges, adding them to the priority queue
        self.log.debug("Computing initial merges")
        for idx, wc in enumerate(self.wrapped_clusters):
            self.log.debug("Computing potential merges for cluster index %d" % idx)
            wc.compute_merges(True)
        
        # greedily apply the least-cost merges
        steps = 0
        while not self.prioritised_merges.empty():
            self.log.debug("Clustering step %d" % (steps+1))
            have_merge = False
            merge: Optional[GreedyAgglomerativeClustering.ClusterMerge] = None
            while not have_merge and not self.prioritised_merges.empty():
                merge = self.prioritised_merges.get()
                if not merge.evaporated:
                    have_merge = True
            if have_merge:
                merge.apply()
            steps += 1
        
        result = filter(lambda wc: not wc.is_merged(), self.wrapped_clusters)
        result = list(map(lambda wc: wc.cluster, result))
        return result

    class WrappedCluster(object):
        """
        Wrapper for clusters which stores additional data required for clustering (internal use only)
        """
        def __init__(self, cluster: "GreedyAgglomerativeClustering.Cluster", idx, clusterer: "GreedyAgglomerativeClustering"):
            self.merged_into_cluster: Optional[GreedyAgglomerativeClustering.WrappedCluster] = None
            self.merges = []
            self.cluster = cluster
            self.idx = idx
            self.clusterer = clusterer

        def is_merged(self) -> bool:
            return self.merged_into_cluster is not None

        def get_cluster_association(self) -> "GreedyAgglomerativeClustering.WrappedCluster":
            """
            Gets the wrapped cluster that this cluster's points have ultimately been merged into (which may be the cluster itself)

            :return: the wrapped cluster this cluster's points are associated with
            """
            if self.merged_into_cluster is None:
                return self
            else:
                return self.merged_into_cluster.get_cluster_association()
            
        def remove_merges(self):
            for merge in self.merges:
                merge.evaporated = True
            self.merges = []

        def compute_merges(self, initial: bool, merged_cluster_indices: Tuple[int, int] = None):
            # add new merges to queue
            wrapped_clusters = self.clusterer.wrapped_clusters
            for item in self.clusterer.mergeCandidateDeterminationStrategy.iter_candidate_indices(self, initial, merged_cluster_indices):
                merge: Optional[GreedyAgglomerativeClustering.ClusterMerge] = None
                if type(item) == int:
                    other_idx = item
                    if other_idx != self.idx:
                        other = wrapped_clusters[other_idx]
                        if not other.is_merged():
                            merge_cost = self.cluster.merge_cost(other.cluster)
                            if not math.isinf(merge_cost):
                                merge = GreedyAgglomerativeClustering.ClusterMerge(self, other, merge_cost)
                else:
                    merge = item
                    assert merge.c1.idx == self.idx
                if merge is not None:
                    merge.c1.merges.append(merge)
                    merge.c2.merges.append(merge)
                    self.clusterer.prioritised_merges.put(merge)

        def __str__(self):
            return "Cluster[idx=%d]" % self.idx
        
    class ClusterMerge(object):
        """
        Represents a potential merge
        """
        log = log.getChild(__qualname__)

        def __init__(self, c1: "GreedyAgglomerativeClustering.WrappedCluster", c2: "GreedyAgglomerativeClustering.WrappedCluster",
                merge_cost):
            self.c1 = c1
            self.c2 = c2
            self.merge_cost = merge_cost
            self.evaporated = False

        def apply(self):
            c1, c2 = self.c1, self.c2
            self.log.debug("Merging %s into %s..." % (str(c1), str(c2)))
            c1.cluster.merge(c2.cluster)
            c2.merged_into_cluster = c1
            c1.remove_merges()
            c2.remove_merges()
            self.log.debug("Computing new merge costs for %s..." % str(c1))
            c1.compute_merges(False, merged_cluster_indices=(c1.idx, c2.idx))
        
        def __lt__(self, other):
            return self.merge_cost < other.merge_cost

    class MergeCandidateDeterminationStrategy(ABC):
        """
        Determines the indices of clusters which should be evaluated with regard to their merge costs
        """
        def __init__(self):
            self.clusterer: Optional["GreedyAgglomerativeClustering"] = None

        def set_clusterer(self, clusterer: "GreedyAgglomerativeClustering"):
            """
            Initialises the clusterer the strategy is applied to
            :param clusterer: the clusterer
            """
            self.clusterer = clusterer

        @abstractmethod
        def iter_candidate_indices(self, wc: "GreedyAgglomerativeClustering.WrappedCluster", initial: bool,
                merged_cluster_indices: Tuple[int, int] = None) -> Iterator[Union[int, "GreedyAgglomerativeClustering.ClusterMerge"]]:
            """
            :param wc: the wrapped cluster: the cluster for which to determine the cluster indices that are to be considered for
                a potential merge
            :param initial: whether we are computing the initial candidates (at the start of the clustering algorithm)
            :param merged_cluster_indices: [for initial=False] the pair of cluster indices that were just joined to form the updated
                cluster wc
            :return: an iterator of cluster indices that should be evaluated as potential merge partners for wc (it may contain the
                index of wc, which will be ignored) or of corresponding ClusterMerge objects
            """
            pass

    class MergeCandidateDeterminationStrategyDefault(MergeCandidateDeterminationStrategy):
        def iter_candidate_indices(self, wc: "GreedyAgglomerativeClustering.WrappedCluster", initial: bool,
                merged_cluster_indices: Tuple[int, int] = None) -> Iterator[Union[int, "GreedyAgglomerativeClustering.ClusterMerge"]]:
            n = len(self.clusterer.wrapped_clusters)
            if initial:
                yield from range(wc.idx + 1, n)
            else:
                yield from range(n)
